"""This module contains the implementation of the BrokerCoinbaseSpot class."""

import time

import requests
import pandas as pd

from pricehub.brokers.broker_abc import BrokerABC

from pricehub.config import TIMEOUT_SEC


class BrokerCoinbaseSpot(BrokerABC):
    """
    Coinbase Spot Broker
    https://docs.cdp.coinbase.com/exchange/reference/exchangerestapi_getproductcandles

    """

    api_url = "https://api.exchange.coinbase.com/products/{symbol}/candles"
    columns = ["Open time", "Low", "High", "Open", "Close", "Volume"]

    interval_map = {
        "1m": 60,
        "5m": 300,
        "15m": 900,
        "1h": 3600,
        "6h": 21600,
        "1d": 86400,
    }
    maximum_data_points = 300

    def fetch_data(self, get_ohlc_params: "GetOhlcParams") -> list:  # type: ignore[name-defined]
        start_time = int(get_ohlc_params.start.timestamp())
        end_time = int(get_ohlc_params.end.timestamp())
        granularity = self.interval_map[get_ohlc_params.interval]

        aggregated_data = []

        while start_time < end_time:
            chunk_end_time = min(start_time + (self.maximum_data_points * granularity), end_time)

            params = {
                "start": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(start_time)),
                "end": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(chunk_end_time)),
                "granularity": granularity,
            }

            url = self.api_url.format(symbol=get_ohlc_params.symbol)
            response = requests.get(url, params=params, timeout=TIMEOUT_SEC)
            data = response.json()

            if not data:
                break

            aggregated_data.extend(data)
            start_time = data[0][0] + granularity

            if chunk_end_time >= end_time:
                break

        return aggregated_data

    def convert_to_dataframe(self, aggregated_data: list) -> pd.DataFrame:
        df = pd.DataFrame(
            aggregated_data,
            columns=["Open time", "Low", "High", "Open", "Close", "Volume"],
        )
        df["Open time"] = pd.to_datetime(df["Open time"], unit="s")
        df.set_index("Open time", inplace=True)
        df.sort_index(inplace=True)

        return df



"""Get OHLC data from a broker"""

import pandas as pd

from pricehub.models import SupportedBroker, Timestamp, GetOhlcParams, Interval


def get_ohlc(
    broker: SupportedBroker, symbol: str, interval: Interval, start: Timestamp, end: Timestamp
) -> pd.DataFrame:
    """
    Get OHLC data from a broker

    Example:

    df = get_ohlc(broker="binance_spot", symbol="BTCUSDT", interval="1h", start="2024-10-01", end="2024-10-02")

    :param interval: Interval of the OHLC data, for example: 1m, 1h, 1d
    :param broker: Name of the broker
    :param symbol: Symbol of the asset
    :param start: Start time of the OHLC data
    :param end: End time of the OHLC data
    :return: OHLC data in a pandas DataFrame
    """
    get_ohlc_params = GetOhlcParams(broker=broker, symbol=symbol, interval=interval, start=start, end=end)
    return get_ohlc_impl(get_ohlc_params)


def get_ohlc_impl(get_ohlc_params: GetOhlcParams) -> pd.DataFrame:
    """
    Get OHLC data from a broker implementation
    :param get_ohlc_params:
    :return: OHLC data in a pandas DataFrame
    """
    broker_class = get_ohlc_params.broker.get_broker_class()
    broker_instance = broker_class()
    return broker_instance.get_ohlc(get_ohlc_params)



from unittest.mock import patch
from pricehub.get_ohlc_impl import get_ohlc, get_ohlc_impl
from pricehub.models import GetOhlcParams


@patch("pricehub.get_ohlc_impl.get_ohlc_impl")
def test_get_ohlc(mock_get_ohlc_impl, get_ohlc_binance_spot_params, get_mock_binance_response_df):
    mock_get_ohlc_impl.return_value = get_mock_binance_response_df

    result = get_ohlc(**get_ohlc_binance_spot_params)

    assert result.equals(get_mock_binance_response_df)
    mock_get_ohlc_impl.assert_called_once_with(GetOhlcParams(**get_ohlc_binance_spot_params))


def test_get_ohlc_impl(get_ohlc_binance_spot_params, get_mock_binance_response_df):
    get_ohlc_params = GetOhlcParams(**get_ohlc_binance_spot_params)
    with patch.object(get_ohlc_params.broker.get_broker_class(), "get_ohlc") as mock_get_ohlc:
        mock_get_ohlc.return_value = get_mock_binance_response_df

        result = get_ohlc_impl(get_ohlc_params)

        assert result.equals(get_mock_binance_response_df)
        mock_get_ohlc.assert_called_once_with(get_ohlc_params)



"""Bybit Spot broker implementation"""

import requests
import pandas as pd
from pricehub.brokers.broker_abc import BrokerABC
from pricehub.config import TIMEOUT_SEC


class BrokerBybitABC(BrokerABC):
    """
    Bybit Spot broker implementation
    https://bybit-exchange.github.io/docs/v5/market/kline
    """

    api_url = "https://api.bybit.com/v5/market/kline"

    interval_map = {
        "1m": "1",
        "3m": "3",
        "5m": "5",
        "15m": "15",
        "30m": "30",
        "1h": "60",
        "2h": "120",
        "4h": "240",
        "6h": "360",
        "12h": "720",
        "1d": "D",
        "1w": "W",
        "1M": "M",
    }

    category = ""
    columns = [
        "Open time",
        "Open",
        "High",
        "Low",
        "Close",
        "Volume",
        "Turnover",
    ]

    maximum_data_points = 200

    def fetch_data(self, get_ohlc_params: "GetOhlcParams") -> list:  # type: ignore[name-defined]
        start_time = int(get_ohlc_params.start.timestamp() * 1000)
        end_time = int(get_ohlc_params.end.timestamp() * 1000)

        params = {
            "category": self.category,
            "symbol": get_ohlc_params.symbol,
            "interval": self.interval_map[get_ohlc_params.interval],
            "start": start_time,
            "end": end_time,
            "limit": self.maximum_data_points,
        }

        aggregated_data = []

        while True:
            response = requests.get(self.api_url, params=params, timeout=TIMEOUT_SEC)
            data = response.json()

            if data["retCode"] != 0:
                raise ValueError(f"Bybit API error: {data['retMsg']}")

            result = data["result"]
            if not result["list"]:
                break

            batch_data = result["list"][::-1]
            aggregated_data.extend(batch_data)

            earliest_timestamp = int(batch_data[0][0])
            params["end"] = earliest_timestamp - 1

            if params["end"] < start_time:
                break

        return aggregated_data

    def convert_to_dataframe(self, aggregated_data: list) -> pd.DataFrame:
        df = pd.DataFrame(
            aggregated_data,
            columns=["Open time", "Open", "High", "Low", "Close", "Volume", "Turnover"],
        )
        df["Open time"] = pd.to_datetime(df["Open time"].astype(int), unit="ms")
        df = df.astype({"Open": float, "High": float, "Low": float, "Close": float, "Volume": float, "Turnover": float})
        df.set_index("Open time", inplace=True)

        df.sort_index(inplace=True)

        return df



"""Bybit Spot broker implementation"""

from pricehub.brokers.broker_bybit_abc import BrokerBybitABC


class BrokerBybitInverse(BrokerBybitABC):
    """
    Bybit Inverse broker implementation
    https://bybit-exchange.github.io/docs/v5/market/kline
    """

    category = "inverse"



"""This module is used to import all the models in the package."""

from pricehub.models.broker import Broker
from pricehub.models.types_common import SupportedBroker, Timestamp, Interval
from pricehub.models.get_ohlc_params import GetOhlcParams

GetOhlcParams.model_rebuild()



"""Binance Spot Broker"""

from pricehub.brokers.broker_binance_abc import BrokerBinanceABC


class BrokerBinanceSpot(BrokerBinanceABC):
    """Binance Spot Broker"""

    api_url = "https://api.binance.com/api/v3/klines"

    interval_map = {
        "1s": "1s",
        "1m": "1m",
        "3m": "3m",
        "5m": "5m",
        "15m": "15m",
        "30m": "30m",
        "1h": "1h",
        "2h": "2h",
        "4h": "4h",
        "6h": "6h",
        "12h": "12h",
        "1d": "1d",
        "1w": "1w",
        "1M": "1M",
    }



"""Common types used in the models of the pricehub app."""

from typing import Literal, Union
import datetime
import pandas as pd


import arrow

SupportedBroker = Literal[
    "binance_spot", "binance_futures", "bybit_spot", "bybit_linear", "bybit_inverse", "coinbase_spot", "okx_spot"
]
Timestamp = Union[int, float, str, arrow.Arrow, pd.Timestamp, datetime.datetime, datetime.date]
Interval = Literal["1s", "1m", "3m", "5m", "15m", "30m", "1h", "2h", "4h", "6h", "8h", "12h", "1d", "3d", "1w", "1M"]



"""Binance Broker ABC Class"""

from abc import ABC

import requests
import pandas as pd

from pricehub.brokers.broker_abc import BrokerABC
from pricehub.config import TIMEOUT_SEC


class BrokerBinanceABC(BrokerABC, ABC):
    """
    Binance Broker  ABC Class
    """

    columns = [
        "Open time",
        "Open",
        "High",
        "Low",
        "Close",
        "Volume",
        "Close time",
        "Quote asset volume",
        "Number of trades",
        "Taker buy base asset volume",
        "Taker buy quote asset volume",
        "Ignore",
    ]

    def fetch_data(self, get_ohlc_params: "GetOhlcParams") -> list:  # type: ignore[name-defined]
        """
        Implement the fetch_data method for Binance
        """

        start_time = int(get_ohlc_params.start.timestamp() * 1000)
        end_time = int(get_ohlc_params.end.timestamp() * 1000)

        aggregated_data: list = []

        while start_time < end_time:
            url = (
                f"{self.api_url}?symbol={get_ohlc_params.symbol}&interval={get_ohlc_params.interval}"
                f"&startTime={start_time}&endTime={end_time}"
            )
            data = requests.get(url, timeout=TIMEOUT_SEC).json()
            if not data:
                break
            aggregated_data.extend(data)
            start_time = data[-1][0] + 1

        return aggregated_data

    def convert_to_dataframe(self, aggregated_data: list) -> pd.DataFrame:
        """
        Implement the convert_to_dataframe method for Binance
        """
        df = pd.DataFrame(aggregated_data, columns=self.columns)
        df = df.astype(float)
        df["Open time"] = pd.to_datetime(df["Open time"], unit="ms")
        df["Close time"] = pd.to_datetime(df["Close time"], unit="ms")
        df.set_index("Open time", inplace=True)

        return df



"""Abstract base class for brokers"""

from abc import ABC, abstractmethod
import pandas as pd


class BrokerABC(ABC):
    """
    Abstract base class for brokers
    """

    @property
    @abstractmethod
    def interval_map(self) -> dict:
        """
        Mapping of intervals supported by the broker.
        :return: dictionary of intervals
        """

    @property
    @abstractmethod
    def columns(self) -> list:
        """
        Columns for the DataFrame.
        :return:
        """

    @property
    @abstractmethod
    def api_url(self) -> str:
        """
        API URL for the broker.
        :return:
        """

    def get_ohlc(self, get_ohlc_params: "GetOhlcParams") -> pd.DataFrame:  # type: ignore[name-defined]
        """
        Get OHLC data from the broker.
        :param get_ohlc_params:
        :return:
        """
        self.validate_interval(get_ohlc_params)
        aggregated_data = self.fetch_data(get_ohlc_params)
        df = self.convert_to_dataframe(aggregated_data)
        return df

    def validate_interval(self, get_ohlc_params: "GetOhlcParams") -> None:  # type: ignore[name-defined]
        """
        Validate the interval for the given broker.
        :param get_ohlc_params:
        :return:
        """
        interval = self.interval_map.get(get_ohlc_params.interval)
        broker_name = self.__class__.__name__
        if not interval:
            raise ValueError(
                f"Interval '{get_ohlc_params.interval}' is not supported by {broker_name}."
                f"Supported intervals: {list(self.interval_map.keys())}"
            )

    @abstractmethod
    def fetch_data(self, get_ohlc_params: "GetOhlcParams") -> list:  # type: ignore[name-defined]
        """
        Fetch data from the broker.
        :param get_ohlc_params:
        :return:
        """

    @abstractmethod
    def convert_to_dataframe(self, aggregated_data: list) -> pd.DataFrame:
        """
        Convert the fetched data to a DataFrame.
        :param aggregated_data:
        :return:
        """



"""Broker Enum class"""

from enum import Enum

from pricehub.brokers.broker_binance_futures import BrokerBinanceFutures
from pricehub.brokers.broker_binance_spot import BrokerBinanceSpot
from pricehub.brokers.broker_bybit_inverse import BrokerBybitInverse
from pricehub.brokers.broker_bybit_linear import BrokerBybitLinear
from pricehub.brokers.broker_bybit_spot import BrokerBybitSpot
from pricehub.brokers.broker_coinbase_spot import BrokerCoinbaseSpot
from pricehub.brokers.broker_okx_spot import BrokerOkxSpot


class Broker(Enum):
    """
    Broker Enum class
    """

    BINANCE_SPOT = "binance_spot"
    BINANCE_FUTURES = "binance_futures"
    BYBIT_SPOT = "bybit_spot"
    BYBIT_LINEAR = "bybit_linear"
    BYBIT_INVERSE = "bybit_inverse"
    COINBASE_SPOT = "coinbase_spot"
    OKX_SPOT = "okx_spot"

    def get_broker_class(self) -> "BrokerABC":  # type: ignore[name-defined]
        """
        Get the broker class for the broker.
        :return:
        """
        broker_classes = {
            Broker.BINANCE_SPOT: BrokerBinanceSpot,
            Broker.BINANCE_FUTURES: BrokerBinanceFutures,
            Broker.BYBIT_SPOT: BrokerBybitSpot,
            Broker.BYBIT_LINEAR: BrokerBybitLinear,
            Broker.BYBIT_INVERSE: BrokerBybitInverse,
            Broker.COINBASE_SPOT: BrokerCoinbaseSpot,
            Broker.OKX_SPOT: BrokerOkxSpot,
        }
        return broker_classes[self]



import requests
import pandas as pd
from typing import List

from pricehub.brokers.broker_abc import BrokerABC
from pricehub.config import TIMEOUT_SEC


class BrokerOkxSpot(BrokerABC):
    """
    OKX Spot Broker (historical)
    https://www.okx.com/docs-v5/en/#rest-api-market-data-get-history-candles
    """

    api_url = "https://www.okx.com/api/v5/market/history-candles"
    columns = ["Open time", "Open", "High", "Low", "Close", "Volume"]

    interval_map = {
        "1m": "1m",
        "5m": "5m",
        "15m": "15m",
        "1h": "1H",
        "6h": "6H",
        "1d": "1D",
    }

    maximum_data_points = 1000

    def fetch_data(self, get_ohlc_params: "GetOhlcParams") -> List[list]:  # type: ignore[name-defined]
        start_ms = int(get_ohlc_params.start.timestamp() * 1000)
        end_ms = int(get_ohlc_params.end.timestamp() * 1000)
        granularity = self.interval_map[get_ohlc_params.interval]

        aggregated = []

        while True:
            params = {
                "instId": get_ohlc_params.symbol,
                "bar": granularity,
                "limit": self.maximum_data_points,
                "after": end_ms,
            }
            resp = requests.get(self.api_url, params=params, timeout=TIMEOUT_SEC)
            resp.raise_for_status()
            data = resp.json().get("data", [])
            if not data:
                break

            for row in data:
                ts = int(row[0])

                if ts < start_ms or ts > end_ms:
                    continue

                aggregated.append(
                    [
                        ts,
                        float(row[1]),
                        float(row[2]),
                        float(row[3]),
                        float(row[4]),
                        float(row[5]),
                    ]
                )

            if int(data[-1][0]) <= start_ms:
                break

            end_ms = int(data[0][0]) * 1000 - 1

        return aggregated[::-1]

    def convert_to_dataframe(self, aggregated_data: list) -> pd.DataFrame:
        df = pd.DataFrame(aggregated_data, columns=self.columns)
        df["Open time"] = pd.to_datetime(df["Open time"], unit="ms", utc=True)
        df.set_index("Open time", inplace=True)
        df.sort_index(inplace=True)
        return df



from setuptools import setup, find_packages


package_name = "pricehub"


setup(
    name=package_name,
    version="0.0.3.14",
    packages=find_packages(),
    author="Evgenii Lazarev",
    author_email="elazarev@gmail.com",
    project_urls={
        "GitHub": "https://github.com/eslazarev",
        "LinkedIn": "https://www.linkedin.com/in/elazarev",
    },
    description="Pricehub: Unified Python Package for Collecting OHLC Prices from Binance, Bybit, and Coinbase APIs into a DataFrame",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "Development Status :: 3 - Alpha",
        "Topic :: Office/Business",
        "Topic :: Office/Business",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
        "Natural Language :: English",
    ],
    url="https://github.com/eslazarev/pricehub",
    install_requires=[
        "pandas",
        "pydantic>=2.0.1",
        "arrow>=1.0.0",
        "requests",
    ],
    python_requires=">=3.8",
    keywords="OHLC OHLCV prices binance bybit coinbase trading  finance quant algotrading cryptocurrency "
    "crypto futures spot broker exchange market quantitative historical candlestick back-testing dataframe",
)



"""GetOhlcParams model."""

from typing import Any

import arrow
from pydantic import BaseModel, field_validator, model_validator, ConfigDict

from pricehub.models import Interval


class GetOhlcParams(BaseModel):
    """
    Get OHLC parameters.
    """

    broker: "Broker"  # type: ignore[name-defined]
    symbol: str
    interval: Interval
    start: arrow.Arrow
    end: arrow.Arrow

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("start", "end", mode="before")
    def convert_to_arrow(cls, value: Any) -> arrow.Arrow:
        """
        Convert the 'start' and 'end' values to Arrow objects.
        :param value:
        :return:
        """
        try:
            return arrow.get(value)
        except Exception as e:
            raise ValueError(f"Invalid date format for value '{value}': {e}") from e

    @model_validator(mode="after")
    def check_start_before_end(self) -> "GetOhlcParams":
        """
        Check that the 'start' date is before the 'end' date.
        :return:
        """
        if self.start > self.end:
            raise ValueError("The 'start' date must be before the 'end' date.")
        return self
